"""Text to speech."""

import os


VOICES = dict(en="Daniel", fi="Satu", nl="Xander")


def say(language: str, text: str) -> None:
    """Say the text in the specified language."""
    voice = VOICES[language]
    text = text.replace("'", r"\'")
    os.system(f"say --voice={voice} {text} &")



"""Labels."""

from typing import cast


class Label(str):
    """Class representing labels for concepts."""
    def __eq__(self, other) -> bool:
        """Ignore hints when determining equality."""
        return self.split(";", maxsplit=1)[0] == other.split(";", maxsplit=1)[0]

    def __ne__(self, other) -> bool:
        """Return whether the labels are not equal."""
        return not self == other

    def spelling_alternatives(self) -> "Labels":
        """Extract the spelling alternatives from the label."""
        return [self.__class__(label) for label in self.split(";", maxsplit=0)[0].split("|")]

    def first_spelling_alternative(self) -> "Label":
        """Extract the first spelling alternative from the label."""
        return self.spelling_alternatives()[0]

    def hint(self) -> str:
        """Return the label hint, if any."""
        return self.split(";")[1] if ";" in self else ""


Labels = list[Label]


def label_factory(string: str | list[str]) -> Labels:
    """Instantiate the labels from a string or list of strings."""
    labels = string if isinstance(string, list) else [string]
    return cast(Labels, [Label(label) for label in labels])



"""Grammar types."""

from typing import Literal


GrammaticalGender = Literal["female", "male"]
GrammaticalNumber = Literal["singular", "plural"]
GrammaticalPerson = Literal["first_person", "second_person", "third_person"]
DegreeOfComparison = Literal["positive_degree", "comparitive_degree", "superlative_degree"]
GrammaticalCategory = Literal[GrammaticalGender, GrammaticalNumber, GrammaticalPerson, DegreeOfComparison]



"""Command to show progress information."""

from rich.table import Table

from toisto.metadata import Language, SUPPORTED_LANGUAGES
from toisto.model import Progress, Quizzes
from toisto.output import console


def show_progress(language: Language, quizzes: Quizzes, progress: Progress) -> None:
    """Show progress."""
    table = Table(title=f"Progress {SUPPORTED_LANGUAGES[language]}")
    table.add_column("Quiz type")
    table.add_column("Question")
    table.add_column("From")
    table.add_column("To")
    table.add_column("Answer(s)")
    table.add_column("Streak", justify="right")
    sorted_quizzes = sorted(quizzes, key=lambda quiz: progress.get_progress(quiz).count, reverse=True)
    for quiz in sorted_quizzes:
        quiz_progress = progress.get_progress(quiz)
        table.add_row(
            quiz.quiz_type.capitalize(),
            quiz.question,
            quiz.question_language,
            quiz.answer_language,
            "\n".join(quiz.answers),
            str(quiz_progress.count) if quiz_progress.count > 0 else ""
        )
    with console.pager():
        console.print(table)



"""Concept classes."""

from __future__ import annotations

from itertools import permutations
from typing import cast, get_args, Iterable, Sequence, Union

from toisto.metadata import Language

from .grammar import GrammaticalCategory
from .label import Labels, label_factory
from .quiz import Quizzes, Quiz, QuizType, quiz_factory, quiz_type_factory


ConceptDict = dict[Language, str | list[str]]
CompositeConceptDict = dict[GrammaticalCategory, Union["CompositeConceptDict", ConceptDict]]


class Concept:
    """Class representing a concept from a topic."""

    def __init__(self, labels: dict[Language, Labels]) -> None:
        self._labels = labels

    def quizzes(self, language: Language, source_language: Language) -> Quizzes:
        """Generate the possible quizzes from the concept and its labels."""
        return (
            quiz_factory(language, source_language, self.labels(language), self.labels(source_language))
        ) if self.has_labels(language, source_language) else []

    def has_labels(self, *languages: Language) -> bool:
        """Return whether the concept has labels for all the specified languages."""
        return all(language in self._labels for language in languages)

    def labels(self, language: Language) -> Labels:
        """Return the labels for the language."""
        return self._labels[language]

    def leaf_concepts(self) -> Iterable[Concept]:
        """Return self as a list of leaf concepts."""
        return [self]

    @classmethod
    def from_dict(cls, concept_dict: ConceptDict) -> Concept:
        """Instantiate a concept from a dict."""
        return cls({language: label_factory(label) for language, label in concept_dict.items()})


ConceptPair = tuple[Concept, Concept]


class CompositeConcept:
    """A concept that consists of multiple other (sub)concepts."""

    def __init__(self, concepts: Sequence[Concept | CompositeConcept], quiz_types: Sequence[QuizType]) -> None:
        self._concepts = concepts
        self._quiz_types = quiz_types

    def quizzes(self, language: Language, source_language: Language) -> Quizzes:
        """Generate the possible quizzes from the concept."""
        result = []
        for concept in self._concepts:
            result.extend(concept.quizzes(language, source_language))
        if not self.has_labels(language):
            return result
        for (concept1, concept2), quiz_type in self.paired_concepts():
            labels1, labels2 = concept1.labels(language), concept2.labels(language)
            quizzes = [
                Quiz(language, language, label1, [label2], quiz_type) for label1, label2 in zip(labels1, labels2)
                if label1 != label2
            ]
            result.extend(quizzes)
        return result

    def has_labels(self, *languages: Language) -> bool:
        """Return whether the concept has labels for all the specified languages."""
        return all(concept.has_labels(*languages) for concept in self._concepts)

    def leaf_concepts(self) -> Iterable[Concept]:
        """Return a list of leaf concepts."""
        for concept in self._concepts:
            for leaf_concept in concept.leaf_concepts():
                yield leaf_concept

    def paired_concepts(self) -> Iterable[tuple[ConceptPair, QuizType]]:
        """Pair the leaf concepts from the composite concepts."""
        leaf_concepts = [concept.leaf_concepts() for concept in self._concepts]
        for concept_group in zip(*leaf_concepts):
            for permutation, quiz_type in zip(permutations(concept_group, r=2), self._quiz_types):
                yield cast(ConceptPair, permutation), quiz_type

    @classmethod
    def from_dict(cls, concept_dict: CompositeConceptDict) -> CompositeConcept:
        """Instantiate a concept from a dict."""
        keys = concept_dict.keys()
        return cls(tuple(concept_factory(concept_dict[key]) for key in keys), quiz_type_factory(tuple(keys)))


def concept_factory(concept_dict: CompositeConceptDict | ConceptDict) -> CompositeConcept | Concept:
    """Create a concept from the concept dict."""
    if set(get_args(GrammaticalCategory)) & set(concept_dict):
        return CompositeConcept.from_dict(cast(CompositeConceptDict, concept_dict))
    return Concept.from_dict(cast(ConceptDict, concept_dict))



"""Model classes."""

from dataclasses import dataclass
from datetime import datetime, timedelta
import math


@dataclass
class QuizProgress:
    """Class to keep track of progress on one quiz."""

    count: int = 0  # The number of consecutive correct guesses
    silence_until: datetime | None = None  # Don't quiz this again until after the datetime

    def update(self, correct: bool) -> None:
        """Update the progress of the quiz."""
        if correct:
            self.count += 1
        else:
            self.count = 0
        self.silence_until = self.__calculate_next_quiz() if self.count > 1 else None

    def is_silenced(self):
        """Return whether the quiz is silenced."""
        return self.silence_until > datetime.now() if self.silence_until else False

    def __calculate_next_quiz(self) -> datetime:
        """Calculate when to quiz again based on the number of correct guesses."""
        # Graph of function below: https://www.desmos.com/calculator/itvdhmh6ex
        max_timedelta = timedelta(days=90)  # How often do we quiz when user has perfect recall?
        slope = 0.25  # How fast do we get to the max timedelta?
        x_shift = -5.6  # Determines the minimum period at count = 2
        y_shift = 1  # Make zero the minimum y value instead of -1
        time_delta = (math.tanh(slope * self.count + x_shift) + y_shift) * (max_timedelta / 2)
        return datetime.now() + time_delta

    def as_dict(self) -> dict[str, int | str]:
        """Return the quiz progress as dict."""
        result: dict[str, int | str] = dict(count=self.count)
        if self.silence_until:
            result["silence_until"] = self.silence_until.isoformat()
        return result

    @classmethod
    def from_dict(cls, quiz_progress_dict: dict[str, int | str]) -> "QuizProgress":
        """Instantiate a quiz progress from a dict."""
        count = int(quiz_progress_dict.get("count", 0))
        silence_until_text = str(quiz_progress_dict.get("silence_until", ""))
        silence_until = datetime.fromisoformat(silence_until_text) if silence_until_text else None
        return cls(count, silence_until)



"""Main module for the application."""

import readline  # pylint: disable=unused-import

from .cli import parser
from .command import practice, show_progress
from .persistence import load_quizzes, load_progress


def main():
    """Main program."""
    namespace = parser.parse_args()
    quizzes = load_quizzes(namespace.language, namespace.source_language, namespace.topic, namespace.topic_file)
    progress = load_progress()
    if namespace.command == "practice":
        practice(quizzes, progress)
    else:
        show_progress(namespace.language, quizzes, progress)



"""Progress model class."""

import random

from .quiz import Quiz, Quizzes
from .quiz_progress import QuizProgress


class Progress:
    """Keep track of progress on quizzes."""
    def __init__(self, progress_dict: dict[str, dict[str, int | str]]) -> None:
        self.__progress_dict = {key: QuizProgress.from_dict(value) for key, value in progress_dict.items()}
        self.__current_quiz: Quiz | None = None

    def update(self, quiz: Quiz, correct: bool) -> None:
        """Update the progress on the quiz."""
        self.__progress_dict.setdefault(str(quiz), QuizProgress()).update(correct)

    def get_progress(self, quiz: Quiz) -> QuizProgress:
        """Return the progress on the quiz."""
        return self.__progress_dict.get(str(quiz), QuizProgress())

    def next_quiz(self, quizzes: Quizzes) -> Quiz | None:
        """Return the next quiz."""
        eligible_quizzes = self.__eligible_quizzes(quizzes)
        self.__current_quiz = random.choice(eligible_quizzes) if eligible_quizzes else None
        return self.__current_quiz

    def __eligible_quizzes(self, quizzes: Quizzes) -> Quizzes:
        """Return eligible quizzes."""
        eligible_quizzes = [quiz for quiz in quizzes if not self.__is_silenced(quiz) and quiz != self.__current_quiz]
        quizzes_with_progress = [quiz for quiz in eligible_quizzes if self.__has_progress(quiz)]
        return eligible_quizzes if len(quizzes_with_progress) < 3 else quizzes_with_progress

    def __is_silenced(self, quiz: Quiz) -> bool:
        """Is the quiz silenced?"""
        return self.get_progress(quiz).is_silenced()

    def __has_progress(self, quiz: Quiz) -> bool:
        """Has the quiz been presented to the user before?"""
        return str(quiz) in self.__progress_dict

    def as_dict(self) -> dict[str, dict[str, int | str]]:
        """Return the progress as dict."""
        return {key: value.as_dict() for key, value in self.__progress_dict.items()}



"""Command-line interface."""

import argparse

from .metadata import SUMMARY, VERSION, TOPICS, SUPPORTED_LANGUAGES


parser = argparse.ArgumentParser(description=SUMMARY)
parser.add_argument("-V", "--version", action="version", version=VERSION)
parser.add_argument(
    "-t", "--topic", action="append", default=[], choices=TOPICS, metavar="{topic}",
    help="topic to use, can be repeated (default: all); available topics: %(choices)s"
)
parser.add_argument(
    "-f", "--topic-file", action="append", default=[], metavar="{topic file}",
    help="topic file to use, can be repeated"
)
parser.add_argument(
    "language", metavar="{language to practice}", choices=SUPPORTED_LANGUAGES.keys(),
    help="language to practice; available languages: %(choices)s"
)
parser.add_argument(
    "source_language", metavar="{your language}", choices=SUPPORTED_LANGUAGES.keys(),
    help="your language; available languages: %(choices)s"
)
parser.add_argument(
    "command", metavar="{command}", choices=["practice", "progress"], default="practice", nargs="?",
    help="command to perform (default: practice); available commands: %(choices)s"
)



"""Quiz classes."""

from dataclasses import dataclass
from itertools import chain
from typing import cast, Literal

from toisto.match import match
from toisto.metadata import Language, SUPPORTED_LANGUAGES

from .grammar import GrammaticalCategory
from .label import Label, Labels


QuizType = Literal[
    "translate",
    "pluralize",
    "singularize",
    "masculinize",
    "feminize",
    "neuterize",
    "give positive degree",
    "give comparitive degree",
    "give superlative degree",
    "give first person",
    "give second person",
    "give third person",
]
INSTRUCTION: dict[QuizType, str] = {
    "translate": "Translate into",
    "pluralize": "Give the [underline]plural[/underline] in",
    "singularize": "Give the [underline]singular[/underline] in",
    "masculinize": "Give the [underline]male[/underline] form in",
    "feminize": "Give the [underline]female[/underline] form in",
    "neuterize": "Give the [underline]neuter[/underline] form in",
    "give positive degree": "Give the [underline]positive degree[/underline] in",
    "give comparitive degree": "Give the [underline]comparitive degree[/underline] in",
    "give superlative degree": "Give the [underline]superlative degree[/underline] in",
    "give first person": "Give the [underline]first person[/underline] in",
    "give second person": "Give the [underline]second person[/underline] in",
    "give third person": "Give the [underline]third person[/underline] in",
}


@dataclass
class Quiz:
    """Class representing a quiz."""
    question_language: Language
    answer_language: Language
    _question: Label
    _answers: Labels
    quiz_type: QuizType = "translate"

    def is_correct(self, guess: Label) -> bool:
        """Return whether the guess is correct."""
        return match(guess, *self.answers)

    @property
    def question(self) -> Label:
        """Return the first spelling alternative of the question."""
        return self._question.first_spelling_alternative()

    @property
    def answer(self) -> Label:
        """Return the first spelling alternative of the first answer."""
        return self._answers[0].first_spelling_alternative()

    @property
    def answers(self) -> Labels:
        """Return all answers."""
        answers = [answer.spelling_alternatives() for answer in self._answers]
        return cast(Labels, list(chain(*answers)))

    def other_answers(self, guess: Label) -> Labels:
        """Return the answers not equal to the guess."""
        assert self.is_correct(guess)
        return [answer for answer in self.answers if not match(guess, answer)]

    def instruction(self) -> str:
        """Generate the quiz instruction."""
        if hint := self._question.hint():
            hint = f" ({hint})"
        return f"{INSTRUCTION[self.quiz_type]} {SUPPORTED_LANGUAGES[self.answer_language]}{hint}"


Quizzes = list[Quiz]


def quiz_factory(  # pylint: disable=too-many-arguments
    language1: Language,
    language2: Language,
    labels1: Labels,
    labels2: Labels,
    quiz_type1: QuizType = "translate",
    quiz_type2: QuizType = "translate"
) -> Quizzes:
    """Create quizzes."""
    return (
        [Quiz(language1, language2, label, labels2, quiz_type1) for label in labels1] +
        [Quiz(language2, language1, label, labels1, quiz_type2) for label in labels2]
    )


def quiz_type_factory(grammatical_categories: tuple[GrammaticalCategory, ...]) -> tuple[QuizType, ...]:
    """Generate the quiz types from the grammatical categories."""
    match grammatical_categories:
        case ("singular", "plural"):
            return ("pluralize", "singularize")
        case ("female", "male", "neuter"):
            return ("masculinize", "neuterize", "feminize", "neuterize", "feminize", "masculinize")
        case ("female", "male"):
            return ("masculinize", "feminize")
        case ("positive_degree", "comparitive_degree", "superlative_degree"):
            return (
                "give comparitive degree", "give superlative degree", "give positive degree",
                "give superlative degree", "give positive degree", "give comparitive degree"
            )
        case ("first_person", "second_person", "third_person"):
            return (
                "give second person", "give third person", "give first person",
                "give third person", "give first person", "give second person"
            )
        case _:
            raise NotImplementedError(f"Don't know how to generate guizzes for {grammatical_categories}")



"""Output for the user."""

from datetime import datetime, timedelta
from typing import NoReturn
import random
import sys

from rich.console import Console
from rich.theme import Theme

from .diff import colored_diff
from .metadata import NAME, VERSION
from .model import Label, Quiz, QuizProgress


theme = Theme({
    "secondary": "grey69",
    "quiz": "medium_purple1",
    "inserted": "bright_green",
    "deleted": "bright_red"
})

console = Console(theme=theme)

WELCOME = f"""👋 Welcome to {NAME} [white not bold]v{VERSION}[/white not bold]!

Practice as many words and phrases as you like, for as long as you like.
Hit Ctrl-C or Ctrl-D to quit.

[secondary]{NAME} tracks how many times you correctly translate words and phrases.
When you correctly translate a word or phrase multiple times in a row,
{NAME} will not quiz you on it for a while. The more correct translations
in a row, the longer words and phrases are silenced.[/secondary]
"""

DONE = f"""👍 Good job. You're done for now. Please come back later or try a different topic.
[secondary]Type `{NAME.lower()} -h` for more information.[/secondary]
"""

TRY_AGAIN = "⚠️  Incorrect. Please try again."

CORRECT = "✅ Correct.\n"

PRAISE = [
    "Nice", "Cool", "Cheers", "Kudos", "Well-done", "Great", "Grand", "Swell", "Bravo", "Great job", "Nailed it",
    "Outstanding", "Excellent", "Awesome", "Terrific", "Brilliant", "Lovely", "Tremendous", "Wow", "Marvellous",
    "Beautiful", "Amazing", "Remarkable", "Magnificent", "Splendid", "Superb", "Wonderful", "Incredible", "Impressive",
    "Sensational", "Fantastic", "Extraordinary", "Sublime", "Spectacular", "Perfect", "Awe-inspiring"]


def format_duration(duration: timedelta) -> str:
    """Format the duration in a human friendly way."""
    if duration.days >= 1.5:
        return f"{duration.days} days"
    if duration.seconds >= 1.5 * 3600:  # 1.5 hours
        hours = round(duration.seconds / 3600)
        return f"{hours} hours"
    minutes = round(duration.seconds / 60)
    return f"{minutes} minutes"


def feedback_correct(guess: Label, quiz: Quiz, quiz_progress: QuizProgress) -> str:
    """Return the feedback about a correct result."""
    text = CORRECT
    if quiz_progress.silence_until:
        silence_duration = format_duration(quiz_progress.silence_until - datetime.now())
        praise = random.choice(PRAISE)
        text += (
            f"[secondary]{praise}! That's {quiz_progress.count} times in a row. "
            f"Skipping this quiz for {silence_duration}.[/secondary]\n"
        )
    if other_answers := quiz.other_answers(guess):
        label = "Another correct answer is" if len(other_answers) == 1 else "Other correct answers are"
        enumerated_answers = ", ".join([f'"{answer}"' for answer in other_answers])
        text += f"""[secondary]{label} {enumerated_answers}.[/secondary]\n"""
    return text


def feedback_incorrect(guess: Label, quiz: Quiz) -> str:
    """Return the feedback about an incorrect result."""
    diff = colored_diff(guess, quiz.answer)
    return f'❌ Incorrect. The correct answer is "{diff}".\n'


def instruction(quiz: Quiz) -> str:
    """Return the instruction for the quiz."""
    return f"[quiz]{quiz.instruction()}:[/quiz]"


def show_error_and_exit(message: str) -> NoReturn:
    """Print the error message to stderr and exit."""
    sys.stderr.write(message)
    sys.exit(2)
